# vim:set fileencoding=utf-8 ft=python ts=8 sw=4 sts=4 et cindent:

# Copyright © 2011, 2012  Fabian Knittel <fabian.knittel@lettink.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.

from __future__ import absolute_import

import paramiko

class AuthError(Exception):
    pass

class FormatError(Exception):
    pass

def retrieve_remote_full_names(connecting_username,
        remote_hostname='rzstud.rz.uni-karlsruhe.de',
        username_regex=None,
        password=None,
        password_query_hook=None,
        feedback_hook=None):
    """Returns the a dictionary of username to full usernames of all users
    matching regex `username_regex` and local to `remote_hostname`.

    Users not matching the `username_regex` are filtered out.

    The query is performed as `connecting_username`.

    Returns ``None`` in case the connection could not be established.

    Uses SSH to connect.

    In case the connection cannot be established without a password and
    `password` is None, `password_query_hook` is called to retrieve a password.
    If the hook is not set or returns ``None``, the connection attempt fails.
    """
    ssh_cmd = SshRemoteCommand(remote_hostname, connecting_username,
                    'getent passwd', password=password,
                    password_query_hook=password_query_hook)
    ssh_cmd.connect()
    if ssh_cmd.stdout is None:
        return None
    ssh_cmd.stdin.close()
    ssh_cmd.stderr.close()
    if feedback_hook is None:
        feedback_hook = lambda x: None

    username_full_name = {}
    for passwd_entry in ssh_cmd.stdout:
        passwd_entries = passwd_entry.strip().split(':')
        if len(passwd_entries) != 7:
            raise FormatError('server answer unexpectedly formatted: "%s"' % \
                    (passwd_entry))
        username = unicode(passwd_entries[0], 'utf-8')
        if username_regex is not None and \
                username_regex.match(username) is None:
            continue
        full_name = unicode(passwd_entries[4], 'utf-8')
        username_full_name[username] = full_name

        num_entries = len(username_full_name)
        if num_entries % 1000 == 0:
            feedback_hook(num_entries)

    feedback_hook(len(username_full_name))
    return username_full_name

class SshRemoteCommand(object):
    """Connects to `hostname` as user `username` and executes `command`.
    Provides access to stdin, stdout and stderr as properties.
    Uses SSH to connect.  In case the connection cannot be established without a
    password, `password_query_hook` is called to retrieve a password.  If the
    hook is not set or returns ``None``, the connection attempt fails.
    """
    def __init__(self, hostname, username, command, password=None,
            password_query_hook=None):
        self.hostname = hostname
        self.username = username
        self.command = command
        self.password = password
        self.password_query_hook = password_query_hook

        self.client = None
        self.stdin = None
        self.stdout = None
        self.stderr = None

    def connect(self):
        self.client = paramiko.SSHClient()
        self.client.load_system_host_keys('/etc/ssh/ssh_known_hosts')
        self.client.load_system_host_keys()
        while True:
            try:
                # Note: Password based connections will always fail once (unless
                #       a password is given from the start), so if that ever
                #       proves to be a bottle-neck, we would probably need to
                #       switch away from the SSHClient convenience interface and
                #       use the underyling layers directly.
                self.client.connect(hostname=self.hostname,
                        username=self.username,
                        password=self.password)
                break
            except (paramiko.SSHException, paramiko.AuthenticationException) \
                    as exc:
                if exc.args[0] not in ('No authentication methods available',
                        'Authentication failed.'):
                    raise
                if self.password_query_hook is not None:
                    self.password = self.password_query_hook()
                else:
                    raise AuthError('Login failed, probably due to missing ' \
                            'password')
                if self.password is None:
                    self.close()
                    return

        self.stdin, self.stdout, self.stderr = self.client.exec_command(
                self.command)

    def close(self):
        if self.client is not None:
            self.client.close()
            self.client = None
            self.stdin = None
            self.stdout = None
            self.stderr = None
